@@ -17,6 +17,7 @@
 
 from __future__ import division, print_function, absolute_import
 
+import sys
 import enum
 import math
 import random
@@ -32,20 +33,22 @@ from functools import reduce
 import attr
 
 from hypothesis.errors import InvalidArgument, ResolutionFailed
-from hypothesis.control import note, assume, cleanup, current_build_context
+from hypothesis.control import note, assume, reject, cleanup, \
+    current_build_context
 from hypothesis._settings import note_deprecation
 from hypothesis.internal.cache import LRUReusedCache
 from hypothesis.searchstrategy import SearchStrategy, check_strategy
-from hypothesis.internal.compat import gcd, ceil, floor, hrange, \
-    text_type, get_type_hints, getfullargspec, implements_iterator
-from hypothesis.internal.floats import next_up, next_down, is_negative, \
-    float_to_int, int_to_float, count_between_floats
+from hypothesis.internal.compat import abc, gcd, ceil, floor, hrange, \
+    string_types, get_type_hints, getfullargspec, typing_root_type, \
+    implements_iterator
+from hypothesis.internal.floats import next_up, float_of, next_down, \
+    is_negative, float_to_int, int_to_float, count_between_floats
 from hypothesis.internal.charmap import as_general_categories
 from hypothesis.internal.cathetus import cathetus
 from hypothesis.internal.renaming import renamed_arguments
 from hypothesis.utils.conventions import infer, not_set
 from hypothesis.internal.reflection import proxies, required_args, \
-    define_function_signature
+    is_typed_named_tuple, define_function_signature
 from hypothesis.internal.validation import check_type, try_convert, \
     check_valid_size, check_valid_bound, check_valid_sizes, \
     check_valid_integer, check_valid_interval, check_valid_magnitude
@@ -77,6 +80,11 @@ try:
 except ImportError:
     pass
 
+try:
+    import numpy
+except ImportError:
+    numpy = None
+
 if False:
     from types import ModuleType  # noqa
     from random import Random  # noqa
@@ -370,8 +378,13 @@ def booleans():
 @cacheable
 @defines_strategy_with_reusable_values
 def floats(
-    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
+    min_value=None,  # type: Real
+    max_value=None,  # type: Real
+    allow_nan=None,  # type: bool
+    allow_infinity=None,  # type: bool
+    width=64,  # type: int
 ):
+    # type: (...) -> SearchStrategy[float]
     """Returns a strategy which generates floats.
 
     - If min_value is not None, all values will be >= min_value.
@@ -384,6 +397,13 @@ def floats(
     Where not explicitly ruled out by the bounds, all of infinity, -infinity
     and NaN are possible values generated by this strategy.
 
+    The width argument specifies the maximum number of bits of precision
+    required to represent the generated float. Valid values are 16, 32, or 64.
+    Passing ``width=32`` will still use the builtin 64-bit ``float`` class,
+    but always for values which can be exactly represented as a 32-bit float.
+    Half-precision floats (``width=16``) are only supported on Python 3.6, or
+    if Numpy is installed.
+
     Examples from this strategy have a complicated and hard to explain
     shrinking behaviour, but it tries to improve "human readability". Finite
     numbers will be preferred to infinity and infinity will be preferred to
@@ -399,14 +419,26 @@ def floats(
                     allow_nan
                 ))
 
+    if width not in (16, 32, 64):
+        raise InvalidArgument(
+            'Got width=%r, but the only valid values are the integers 16, '
+            '32, and 64.' % (width,)
+        )
+    if width == 16 and sys.version_info[:2] < (3, 6) and numpy is None:
+        raise InvalidArgument(  # pragma: no cover
+            'width=16 requires either Numpy, or Python >= 3.6'
+        )
+
     check_valid_bound(min_value, 'min_value')
     check_valid_bound(max_value, 'max_value')
 
     min_arg, max_arg = min_value, max_value
     if min_value is not None:
-        min_value = float(min_value)
+        min_value = float_of(min_value, width)
+        assert isinstance(min_value, float)
     if max_value is not None:
-        max_value = float(max_value)
+        max_value = float_of(max_value, width)
+        assert isinstance(max_value, float)
 
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
     if min_value == float(u'-inf'):
@@ -414,16 +446,17 @@ def floats(
     if max_value == float(u'inf'):
         max_value = None
 
-    if min_value is not None and min_value < min_arg:
-        min_value = next_up(min_value)
-        assert min_value > min_arg
-    if max_value is not None and max_value > max_arg:
-        max_value = next_down(max_value)
-        assert max_value < max_arg
-    if None not in (min_value, max_value) and min_value > max_value:
+    if min_value is not None and min_arg is not None and min_value < min_arg:
+        min_value = next_up(min_value, width)
+        assert min_value > min_arg  # type: ignore
+    if max_value is not None and max_arg is not None and max_value > max_arg:
+        max_value = next_down(max_value, width)
+        assert max_value < max_arg  # type: ignore
+    if min_value is not None and max_value is not None and \
+            min_value > max_value:
         raise InvalidArgument(
-            'There are no floating-point values between min_value=%r and '
-            'max_value=%r' % (min_arg, max_arg))
+            'There are no %s-bit floating-point values between min_value=%r '
+            'and max_value=%r' % (width, min_arg, max_arg))
 
     if allow_infinity is None:
         allow_infinity = bool(min_value is None or max_value is None)
@@ -436,61 +469,71 @@ def floats(
                 ))
 
     if min_value is None and max_value is None:
-        return FloatStrategy(
+        result = FloatStrategy(
             allow_infinity=allow_infinity, allow_nan=allow_nan,
-        )
+        )  # type: SearchStrategy[float]
     elif min_value is not None and max_value is not None:
         if min_value == max_value:
-            return just(min_value)
+            assert isinstance(min_value, float)
+            result = just(min_value)
         elif is_negative(min_value):
             if is_negative(max_value):
-                return floats(min_value=-max_value, max_value=-min_value).map(
-                    operator.neg
-                )
+                result = floats(min_value=-max_value, max_value=-min_value)\
+                    .map(operator.neg)
             else:
-                return floats(min_value=0.0, max_value=max_value) | floats(
+                result = floats(min_value=0.0, max_value=max_value) | floats(
                     min_value=0.0, max_value=-min_value).map(operator.neg)
         elif count_between_floats(min_value, max_value) > 1000:
-            return FixedBoundedFloatStrategy(
+            result = FixedBoundedFloatStrategy(
                 lower_bound=min_value, upper_bound=max_value
             )
         else:
-            ub_int = float_to_int(max_value)
-            lb_int = float_to_int(min_value)
+            ub_int = float_to_int(max_value, width)
+            lb_int = float_to_int(min_value, width)
             assert lb_int <= ub_int
-            return integers(min_value=lb_int, max_value=ub_int).map(
-                int_to_float
+            result = integers(min_value=lb_int, max_value=ub_int).map(
+                lambda x: int_to_float(x, width)
             )
     elif min_value is not None:
+        assert isinstance(min_value, float)
         if min_value < 0:
             result = floats(
-                min_value=0.0
+                min_value=0.0, allow_infinity=allow_infinity
             ) | floats(min_value=min_value, max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
-                    lambda x: assume(not math.isnan(x)) and min_value + abs(x)
+                    lambda x: assume(not math.isnan(x)) and
+                    min_value + abs(x)  # type: ignore
                 )
             )
         if min_value == 0 and not is_negative(min_value):
             result = result.filter(lambda x: math.copysign(1.0, x) == 1)
-        return result
     else:
-        assert max_value is not None
+        assert isinstance(max_value, float)
         if max_value > 0:
             result = floats(
                 min_value=0.0,
                 max_value=max_value,
-            ) | floats(max_value=-0.0)
+            ) | floats(max_value=-0.0, allow_infinity=allow_infinity)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
-                    lambda x: assume(not math.isnan(x)) and max_value - abs(x)
+                    lambda x: assume(not math.isnan(x)) and
+                    max_value - abs(x)  # type: ignore
                 )
             )
         if max_value == 0 and is_negative(max_value):
             result = result.filter(is_negative)
-        return result
+
+    if width < 64:
+        def downcast(x):
+            try:
+                return float_of(x, width)
+            except OverflowError:
+                reject()
+        return result.map(downcast)
+    return result
 
 
 @cacheable
@@ -528,7 +571,7 @@ def sampled_from(elements):
 def sampled_from(elements):
     """Returns a strategy which generates any value present in ``elements``.
 
-    Note that as with :func:`~hypotheses.strategies.just`, values will not be
+    Note that as with :func:`~hypothesis.strategies.just`, values will not be
     copied and thus you should be careful of using mutable data.
 
     ``sampled_from`` supports ordered collections, as well as
@@ -574,12 +617,13 @@ def lists(
     Hypothesis 1.x mean we no longer needed this hint to generate useful data.
 
     If unique is True (or something that evaluates to True), we compare direct
-    object equality, as if unique_by was `lambda x: x`. This comparison only
+    object equality, as if unique_by was ``lambda x: x``. This comparison only
     works for hashable types.
 
     if unique_by is not None it must be a function returning a hashable type
     when given a value drawn from elements. The resulting list will satisfy the
-    condition that for i != j, unique_by(result[i]) != unique_by(result[j]).
+    condition that for ``i`` != ``j``, ``unique_by(result[i])`` !=
+    ``unique_by(result[j])``.
 
     Examples from this strategy shrink by trying to remove elements from the
     list, and by shrinking each individual element of the list.
@@ -826,7 +870,7 @@ def characters(
       exception.
 
     The ``_codepoint`` arguments must be integers between zero and
-    :obj:`python:sys.max_unicode`.  The ``_characters`` arguments must be
+    :obj:`python:sys.maxunicode`.  The ``_characters`` arguments must be
     collections of length-one unicode strings, such as a unicode string.
 
     The ``_categories`` arguments must be used to specify either the
@@ -926,7 +970,31 @@ def text(
     elif isinstance(alphabet, SearchStrategy):
         char_strategy = alphabet
     else:
-        char_strategy = sampled_from(list(map(text_type, alphabet)))
+        if not isinstance(alphabet, abc.Sequence):
+            note_deprecation(
+                'alphabet must be an ordered sequence, or tests may be '
+                'flaky and shrinking weaker, but a %r is not a type of '
+                'sequence.  This will be an error in future.'
+                % (type(alphabet),)
+            )
+        alphabet = list(alphabet)
+        non_string = [c for c in alphabet if not isinstance(c, string_types)]
+        if non_string:
+            note_deprecation(
+                'The following elements in alphabet are not unicode '
+                'strings, which will be an error in future:  %r'
+                % (non_string,)
+            )
+            alphabet = [str(c) for c in alphabet]
+        not_one_char = [c for c in alphabet
+                        if isinstance(c, string_types) and len(c) != 1]
+        if not_one_char:
+            note_deprecation(
+                'The following elements in alphabet are not of length '
+                'one, which leads to violation of size constraints and '
+                'will be an error in future:  %r' % (not_one_char,)
+            )
+        char_strategy = sampled_from(alphabet)
     return StringStrategy(lists(
         char_strategy, min_size=min_size, max_size=max_size
     ))
@@ -934,17 +1002,17 @@ def text(
 
 @cacheable
 @defines_strategy
-def from_regex(regex):
-    # type: (Union[AnyStr, Pattern[AnyStr]]) -> SearchStrategy[AnyStr]
+def from_regex(regex, fullmatch=False):
+    # type: (Union[AnyStr, Pattern[AnyStr]], bool) -> SearchStrategy[AnyStr]
     """Generates strings that contain a match for the given regex (i.e. ones
-    for which :func:`re.search` will return a non-None result).
+    for which :func:`python:re.search` will return a non-None result).
 
     ``regex`` may be a pattern or :func:`compiled regex <python:re.compile>`.
     Both byte-strings and unicode strings are supported, and will generate
     examples of the same type.
 
-    You can use regex flags such as :const:`re.IGNORECASE`, :const:`re.DOTALL`
-    or :const:`re.UNICODE` to control generation. Flags can be passed either
+    You can use regex flags such as :obj:`python:re.IGNORECASE` or
+    :obj:`python:re.DOTALL` to control generation. Flags can be passed either
     in compiled regex or inside the pattern with a ``(?iLmsux)`` group.
 
     Some regular expressions are only partly supported - the underlying
@@ -957,14 +1025,17 @@ def from_regex(regex):
     boundary markers. So e.g. ``r"\\A.\\Z"`` will return a single character
     string, while ``"."`` will return any string, and ``r"\\A.$"`` will return
     a single character optionally followed by a ``"\\n"``.
+    Alternatively, passing ``fullmatch=True`` will ensure that the whole
+    string is a match, as if you had used the ``\\A`` and ``\\Z`` markers.
 
     Examples from this strategy shrink towards shorter strings and lower
-    character values.
+    character values, with exact behaviour that may depend on the pattern.
     """
+    check_type(bool, fullmatch, 'fullmatch')
     # TODO: We would like to move this to the top level, but pending some major
     # refactoring it's hard to do without creating circular imports.
     from hypothesis.searchstrategy.regex import regex_strategy
-    return regex_strategy(regex)
+    return regex_strategy(regex, fullmatch)
 
 
 @cacheable
@@ -1100,8 +1171,14 @@ def builds(
             from hypothesis.searchstrategy.attrs import from_attrs
             return from_attrs(target, args, kwargs, required | to_infer)
         # Otherwise, try using type hints
-        hints = get_type_hints(
-            target.__init__ if isclass(target) else target)
+        if isclass(target):
+            if is_typed_named_tuple(target):
+                # Special handling for typing.NamedTuple
+                hints = target._field_types
+            else:
+                hints = get_type_hints(target.__init__)
+        else:
+            hints = get_type_hints(target)
         if to_infer - set(hints):
             raise InvalidArgument(
                 'passed infer for %s, but there is no type annotation'
@@ -1158,7 +1235,6 @@ def from_type(thing):
     from hypothesis.searchstrategy import types
 
     if typing is not None:  # pragma: no branch
-        fr = typing._ForwardRef
         if not isinstance(thing, type):
             # At runtime, `typing.NewType` returns an identity function rather
             # than an actual type, but we can check that for a possible match
@@ -1175,12 +1251,13 @@ def from_type(thing):
                 return one_of([from_type(t) for t in args])
         # We can't resolve forward references, and under Python 3.5 (only)
         # a forward reference is an instance of type.  Hence, explicit check:
-        elif type(thing) == fr:  # pragma: no cover
+        elif hasattr(typing, '_ForwardRef') and \
+                type(thing) == typing._ForwardRef:  # pragma: no cover
             raise ResolutionFailed(
                 'thing=%s cannot be resolved.  Upgrading to python>=3.6 may '
                 'fix this problem via improvements to the typing module.'
                 % (thing,))
-    if not isinstance(thing, type):
+    if not types.is_a_type(thing):
         raise InvalidArgument('thing=%s must be a type' % (thing,))
     # Now that we know `thing` is a type, the first step is to check for an
     # explicitly registered strategy.  This is the best (and hopefully most
@@ -1190,27 +1267,27 @@ def from_type(thing):
     if thing in types._global_type_lookup:
         strategy = types._global_type_lookup[thing]
         if not isinstance(strategy, SearchStrategy):
-            strategy = strategy(thing)
-        if strategy.is_empty:
+            strategy = strategy(thing)  # type: ignore
+        if strategy.is_empty:  # type: ignore
             raise ResolutionFailed(
                 'Error: %r resolved to an empty strategy' % (thing,))
-        return strategy
+        return strategy  # type: ignore
     # If there's no explicitly registered strategy, maybe a subtype of thing
     # is registered - if so, we can resolve it to the subclass strategy.
     # We'll start by checking if thing is from from the typing module,
     # because there are several special cases that don't play well with
     # subclass and instance checks.
     if typing is not None:  # pragma: no branch
-        if isinstance(thing, typing.TypingMeta):
+        if isinstance(thing, typing_root_type):
             return types.from_typing_type(thing)
     # If it's not from the typing module, we get all registered types that are
     # a subclass of `thing` and are not themselves a subtype of any other such
     # type.  For example, `Number -> integers() | floats()`, but bools() is
     # not included because bool is a subclass of int as well as Number.
     strategies = [
-        v if isinstance(v, SearchStrategy) else v(thing)
+        v if isinstance(v, SearchStrategy) else v(thing)  # type: ignore
         for k, v in types._global_type_lookup.items()
-        if issubclass(k, thing) and sum(
+        if isinstance(k, type) and issubclass(k, thing) and sum(
             types.try_issubclass(k, typ) for typ in types._global_type_lookup
         ) == 1
     ]
@@ -1223,25 +1300,21 @@ def from_type(thing):
         return one_of(strategies)
     # If we don't have a strategy registered for this type or any subtype, we
     # may be able to fall back on type annotations.
-    # Types created via typing.NamedTuple use a custom attribute instead -
-    # but we can still use builds(), if we work out the right kwargs.
-    if issubclass(thing, tuple) and hasattr(thing, '_fields') \
-            and hasattr(thing, '_field_types'):
-        kwargs = {k: from_type(thing._field_types[k]) for k in thing._fields}
-        return builds(thing, **kwargs)
     if issubclass(thing, enum.Enum):
-        assert len(thing), repr(thing) + ' has no members to sample'
         return sampled_from(thing)
-    # If the constructor has an annotation for all required arguments,
-    # or the class was defined with attrs, builds(thing) will work.
-    # (unless inference fails for an argument, but that's the same above too)
+    # If we know that builds(thing) will fail, give a better error message
     required = required_args(thing)
-    if not required or required.issubset(get_type_hints(thing.__init__)) or \
-            attr.has(thing):
-        return builds(thing)
-    # We have utterly failed, and might as well say so now.
-    raise ResolutionFailed('Could not resolve %r to a strategy; consider '
-                           'using register_type_strategy' % (thing,))
+    if not any([
+        not required,
+        required.issubset(get_type_hints(thing.__init__)),
+        attr.has(thing),
+        # NamedTuples are weird enough that we need a specific check for them.
+        is_typed_named_tuple(thing),
+    ]):
+        raise ResolutionFailed('Could not resolve %r to a strategy; consider '
+                               'using register_type_strategy' % (thing,))
+    # Finally, try to build an instance by calling the type object
+    return builds(thing)
 
 
 @cacheable
@@ -1391,7 +1464,7 @@ def decimals(
     places=None,  # type: int
 ):
     # type: (...) -> SearchStrategy[Decimal]
-    """Generates instances of :class:`decimals.Decimal`, which may be:
+    """Generates instances of :class:`python:decimal.Decimal`, which may be:
 
     - A finite rational number, between ``min_value`` and ``max_value``.
     - Not a Number, if ``allow_nan`` is True.  None means "allow NaN, unless
@@ -1502,11 +1575,11 @@ class PermutationStrategy(SearchStrategy):
         self.values = values
 
     def do_draw(self, data):
-        # Reversed Fisher-Yates shuffle. Reverse order so that it shrinks
-        # propertly: This way we prefer things that are lexicographically
-        # closer to the identity.
+        # Reversed Fisher-Yates shuffle: swap each element with itself or with
+        # a later element.  This shrinks i==j for each element, i.e. to no
+        # change.  We don't consider the last element as it's always a no-op.
         result = list(self.values)
-        for i in hrange(len(result)):
+        for i in hrange(len(result) - 1):
             j = integer_range(data, i, len(result) - 1)
             result[i], result[j] = result[j], result[i]
         return result
@@ -1543,8 +1616,8 @@ def datetimes(
     # type: (...) -> SearchStrategy[dt.datetime]
     """A strategy for generating datetimes, which may be timezone-aware.
 
-    This strategy works by drawing a naive datetime between ``min_datetime``
-    and ``max_datetime``, which must both be naive (have no timezone).
+    This strategy works by drawing a naive datetime between ``min_value``
+    and ``max_value``, which must both be naive (have no timezone).
 
     ``timezones`` must be a strategy that generates
     :class:`~python:datetime.tzinfo` objects (or None,
@@ -1609,7 +1682,7 @@ def dates(
     min_date=None, max_date=None,
 ):
     # type: (dt.date, dt.date, dt.date, dt.date) -> SearchStrategy[dt.date]
-    """A strategy for dates between ``min_date`` and ``max_date``.
+    """A strategy for dates between ``min_value`` and ``max_value``.
 
     Examples from this strategy shrink towards January 1st 2000.
     """
@@ -1634,7 +1707,7 @@ def times(
     max_time=None,  # type: dt.time
 ):
     # type: (...) -> SearchStrategy[dt.time]
-    """A strategy for times between ``min_time`` and ``max_time``.
+    """A strategy for times between ``min_value`` and ``max_value``.
 
     The ``timezones`` argument is handled as for :py:func:`datetimes`.
 
@@ -1686,13 +1759,7 @@ class CompositeStrategy(SearchStrategy):
         self.kwargs = kwargs
 
     def do_draw(self, data):
-        first_draw = [True]
-
-        def draw(strategy):
-            first_draw[0] = False
-            return data.draw(strategy)
-
-        return self.definition(draw, *self.args, **self.kwargs)
+        return self.definition(data.draw, *self.args, **self.kwargs)
 
     def calc_label(self):
         return self.__label
@@ -1764,8 +1831,8 @@ def complex_numbers(min_magnitude=0, max_magnitude=None,
 
     If you need to generate complex numbers with particular real and
     imaginary parts or relationships between parts, consider using
-    `builds(complex, ...) <hypothesis.strategies.builds>` or
-    `@composite <hypothesis.strategies.composite>` respectively.
+    :func:`builds(complex, ...) <hypothesis.strategies.builds>` or
+    :func:`@composite <hypothesis.strategies.composite>` respectively.
     """
     check_valid_magnitude(min_magnitude, 'min_magnitude')
     check_valid_magnitude(max_magnitude, 'max_magnitude')
@@ -2025,8 +2092,11 @@ def data():
     return DataStrategy()
 
 
-def register_type_strategy(custom_type, strategy):
-    # type: (type, Union[type, Callable[[type], SearchStrategy]]) -> None
+def register_type_strategy(
+    custom_type,  # type: type
+    strategy,  # type: Union[SearchStrategy, Callable[[type], SearchStrategy]]
+):
+    # type: (...) -> None
     """Add an entry to the global type-to-strategy lookup.
 
     This lookup is used in :func:`~hypothesis.strategies.builds` and
@@ -2044,7 +2114,7 @@ def register_type_strategy(custom_type, strategy):
     # TODO: We would like to move this to the top level, but pending some major
     # refactoring it's hard to do without creating circular imports.
     from hypothesis.searchstrategy import types
-    if not isinstance(custom_type, type):
+    if not types.is_a_type(custom_type):
         raise InvalidArgument('custom_type=%r must be a type')
     elif not (isinstance(strategy, SearchStrategy) or callable(strategy)):
         raise InvalidArgument(
@@ -2094,12 +2164,11 @@ def deferred(definition):
 @defines_strategy_with_reusable_values
 def emails():
     """A strategy for generating email addresses as unicode strings. The
-    address format is specific in :rfc:`5322#section-3.4.1`. Values shrink
+    address format is specified in :rfc:`5322#section-3.4.1`. Values shrink
     towards shorter local-parts and host domains.
 
     This strategy is useful for generating "user data" for tests, as
-    mishandling of email addresses is a common source of bugs. Future
-    updates will generate more complicated addresses allowed by the RFC.
+    mishandling of email addresses is a common source of bugs.
     """
     from hypothesis.provisional import domains
     local_chars = string.ascii_letters + string.digits + "!#$%&'*+-/=^_`{|}~"
